import bcrypt from "bcryptjs";
import jwt from "jsonwebtoken";

import { Usuario, UsuarioDoc, UsuarioRepository } from "../models/Usuario.js";
import { AuthSchemas } from "../schemas/authSchemas.js";
import { TokenPayload } from "../types/express/index.js";
import ControllerBase, { ParsedSchema, route } from "./ControllerBase.js";
import { APIError } from "../middlewares/errorHandler.js";
import { getResponseSchema } from "../utils/messages.js";
import { z } from "zod";
import { decodeJWTToken } from "../utils/decodeToken.js";

export default class AuthController extends ControllerBase {
    public repo: UsuarioRepository;

    constructor(
        repo = UsuarioRepository.getInstance()
    ) {
        super("Auth");
        this.repo = repo;
    }

    async _verificarLogin(_email: string | UsuarioDoc | null, senha: string) {
        let usuario: UsuarioDoc | null = null;

        if (!_email) throw new APIError("Email e/ou senha incorretos", 400);
        if(typeof _email === "string") {
            usuario = await this.repo.obterUsuarioPorEmail(_email, {selectSenha: true});
            if (!usuario) throw new APIError("Email e/ou senha incorretos", 400);
        } else {
            usuario = _email;
        }

        if(usuario.banido_ate && usuario.banido_ate > new Date()) throw new APIError("Este usuário está banido", 400);

        const hashMatched = await bcrypt.compare(senha, usuario.senha!);
        if(!hashMatched) throw new APIError("Email e/ou senha incorretos", 400);

        return usuario;
    }

    /**
     * curl -X POST http://localhost:3000/auth \
        -H "Content-Type: application/json" \
        -d '{"email": "Arnoldo91@yahoo.com", "senha": "ABCDabcd1234"}' 
     */
    //static fazerLogin: RequestHandler = async (req, res, next) => {
    @route({
        description: "Fazer login na aplicação",
        method: "post",
        route: "/auth",
        schema: AuthSchemas.fazerLogin,
        response: {code: 200, json: getResponseSchema(z.object({
            token: z.string(),
            payload: AuthSchemas._tokenPayload
        }))}
    })
    async fazerLogin({body}: ParsedSchema<typeof AuthSchemas.fazerLogin>) {
        const usuario = await this._verificarLogin(body.email, body.senha);
        const JWTToken = jwt.sign(
            {
                username: usuario.username,
                email_confirmado: usuario.email_confirmado
            }, 
            process.env.JWT_SECRET!, 
            {
                expiresIn: process.env.JWT_EXPIRES_IN! as any,
                issuer: "AVALIADOR"
            }
        );

        const decoded = decodeJWTToken(JWTToken);

        return {code: 200, json: { 
            token: JWTToken,
            payload: decoded
        }};
    }

    // Middleware já verificou se o token JWT é válido
    @route({
        auth: "require",
        description: "Verificar se o token JWT é válido, e então retorna o payload",
        method: "post",
        route: "/auth/check",
        schema: {},
        response: {code: 200, json: getResponseSchema(z.object({
            payload: AuthSchemas._tokenPayload
        }))}
    })
    async loginCheck(dados: {}, {token}: {token: TokenPayload}) {
        return {code: 200, json: {
            payload: token
        }};
    }

    @route({
        auth: "require",
        description: "Trocar a própria senha, precisa estar logado",
        method: "post",
        route: "/auth/redefinir-senha",
        schema: AuthSchemas.redefinirSenha,
        response: {code: 200, json: getResponseSchema(z.object({
            message: z.string()
        }))}
    })
    async redefinirSenha({body} : ParsedSchema<typeof AuthSchemas.redefinirSenha>, {token}: {token: TokenPayload}) {
        // 1. verificar se a senha antiga está correta
        let usuario = await this.repo.obterUsuarioPorUsername(token!.username, {selectSenha: true});
        usuario = await this._verificarLogin(usuario, body.senha);

        // 2. gerar novo hash de senha
        const senhaBcrypt = await bcrypt.hash(body.senha_nova, await bcrypt.genSalt());
        usuario!.senha = senhaBcrypt;
        await usuario!.save();

        return {code: 200, json: { 
            message: "Senha redefinida com sucesso" 
        }};
    }
}