import { CodigoInteractor } from "../interactors/codigoInteractor.js";
import { ProblemaRepository } from "../models/Problema.js";
import { CodigoSchemas } from "../schemas/codigoSchemas.js";
import { APIError } from "../middlewares/errorHandler.js";
import ControllerBase, { ParsedSchema, route } from "./ControllerBase.js";
import { getPaginatedResponseSchema, getResponseSchema } from "../utils/messages.js";

export default class CodigoControler extends ControllerBase {
    public interactor: CodigoInteractor;
    public problemaRepo: ProblemaRepository;

    constructor(
        codigoInteractor: CodigoInteractor,
        problemaRepo = ProblemaRepository.getInstance()
    ) {
        super("Codigo");
        this.interactor = codigoInteractor;
        this.problemaRepo = problemaRepo;
    }

    static async getInstance() {
        return new CodigoControler(await CodigoInteractor.getInstance());
    }

    @route({
        summary: "Enviar um código para ser avaliado",
        description: `<p> 
            O código é enviado para avaliação, que acontecerá de forma assíncrona, o
            status pode ser acompanhado pela rota /codigos/:id que terá a situação atual
        </p>`,
        method: "post",
        route: "/problemas/:sigla/codigo",
        schema: CodigoSchemas.enviarCodigo,
        response: {code: 201, json: getResponseSchema(CodigoSchemas.codigoGET)},
    })
    async enviarCodigo({params, body}: ParsedSchema<typeof CodigoSchemas.enviarCodigo>) {
        const problemaObj = await this.problemaRepo.obterPorSigla(params.sigla, true);
        if(!problemaObj)
            throw new APIError("Problema não encontrado", 404);

        const codigoObj = await this.interactor.enviarCodigo(problemaObj, body.codigo, body.linguagem);

        return {code: 201, json: codigoObj};
    }

    @route({
        method: "get",
        route: "/codigos/:id",
        schema: CodigoSchemas.codigoPorId,
        response: {code: 200, json: getResponseSchema(CodigoSchemas.codigoGET)}
    })
    async codigoPorId({params}: ParsedSchema<typeof CodigoSchemas.codigoPorId>) {
        const codigo = await this.interactor.repo.findById(params.id);
        if(!codigo)
            throw new APIError("Código não encontrado", 404);

        return {code: 200, json: codigo};
    }

    @route({
        method: "get",
        route: "/codigos",
        schema: CodigoSchemas.meusCodigos,
        response: {code: 200, json: getPaginatedResponseSchema(CodigoSchemas.codigoGET)}
    })
    async meusCodigos({query}: ParsedSchema<typeof CodigoSchemas.meusCodigos>) {
        const result = await this.interactor.repo.listar(query);

        return {code: 200, json: result};
    }
}