import { RequestHandler, Router } from "express";
import { AnyZodObject, z, ZodEffects, ZodSchema } from "zod";
import { ControllerResp, sendResponse } from "../utils/messages.js";
import { APILocalsToken, APILocals, TokenPayload } from "../types/express/index.js";
import { APIError } from "../middlewares/errorHandler.js";

type ZodObjectWithEffect = AnyZodObject | ZodEffects<ZodObjectWithEffect, any, any>;

export type ControllerSchema = {
    query?: ZodObjectWithEffect;
    body?: ZodObjectWithEffect;
    params?: ZodObjectWithEffect;
};

export type ParsedSchema<T extends ControllerSchema> = {
    query: T["query"] extends AnyZodObject ? z.infer<T["query"]> : undefined;
    body: T["body"] extends AnyZodObject ? z.infer<T["body"]> : undefined;
    params: T["params"] extends AnyZodObject ? z.infer<T["params"]> : undefined;
};

export type ResponseSchema = {
    code: number;
    json?: ZodObjectWithEffect;
};

export type ParsedRequestHandler<T extends ControllerSchema> = (
    parsed: ParsedSchema<T>,
    locals: APILocals
) => Promise<{code: number, json?: ControllerResp}>;

export type ParsedAutenticatedRequestHandler<T extends ControllerSchema> = (
    parsed: ParsedSchema<T>,
    locals: APILocalsToken
) => Promise<{code: number, json?: ControllerResp}>;

export type AuthType = "require" | "optional";    
export type RouteDefinition<T extends ControllerSchema, Auth extends AuthType> = {
    method: "get" | "post" | "patch" | "put" | "delete";
    route: string;
    schema: T;
    handler: RequestHandler;

    /**
     * Define se a rota requer autenticação, e qual o nível de autenticação necessário.
     * - "require": decodifica o token e produz erro se não for válido ou se não for informado
     * - "optional": decodifica o token se presente, mas não produz erro se não foi informado
     * - undefined: pula verificação do token
     */
    auth?: Auth;

    // Opcionalmente para documentação:
    summary?: string;
    description?: string;
    response?: ResponseSchema;
};

type RequestHandlerFromDefinition<T extends ControllerSchema, K extends AuthType> = 
    K extends "require" ? ParsedAutenticatedRequestHandler<T> : ParsedRequestHandler<T>;

/**
 * Decorator para definir rotas em um controller,
 * basicamente é um wrapper para o método handler do controller, que adiciona a rota à lista de rotas do controller.
 * 
 * Referências sobre decorators:
 * - https://www.typescriptlang.org/docs/handbook/release-notes/typescript-5-0.html
 * - https://devblogs.microsoft.com/typescript/announcing-typescript-5-0-beta/
 * - https://2ality.com/2022/10/javascript-decorators.html
 */
export function route<T extends ControllerSchema, K extends AuthType>(definition: Omit<RouteDefinition<T,K>, "handler">) {
    return function routeDecorator(
        originalMethod: RequestHandlerFromDefinition<T, K>,
        context: ClassMethodDecoratorContext<ControllerBase, RequestHandlerFromDefinition<T, K>>) {
        const methodName = context.name;
        if (context.private) {
            throw new Error(`'bound' cannot decorate private properties like ${methodName as string}.`);
        }
        context.addInitializer(function(this: ControllerBase) {
            const that: any = this; // sabemos o que estamos fazendo...
            that[methodName] = that[methodName].bind(this);
            this.handler(definition, that[methodName]);
        });
    }
}

export default abstract class ControllerBase {
    declare routes: RouteDefinition<ControllerSchema,AuthType>[];
    name: string; 

    constructor(name: string) {
        if(!this.routes) this.routes = [];

        this.name = name;
    }

    handler<T extends ControllerSchema, K extends AuthType>(definition: Omit<RouteDefinition<T,K>, "handler">, fn: RequestHandlerFromDefinition<T, K>): RequestHandlerFromDefinition<T, K> {
        const handlerFn: RequestHandler = async (req, res, next) => {
            try {
                const schema = definition.schema;
                const parsed: ParsedSchema<T> = {
                    query: schema.query ? schema.query.parse(req.query) : undefined,
                    body: schema.body ? schema.body.parse(req.body) : undefined,
                    params: schema.params ? schema.params.parse(req.params) : undefined,
                };

                if(definition.auth === "require" && !res.locals.token) {
                    throw new APIError("Token não encontrado", 500);
                }

                const {code, json} = await fn(parsed, res.locals as APILocalsToken);
                sendResponse(res, code, json);                
            } catch (e) {
                next(e);
            }
        };
        
        const routeDeclaration: RouteDefinition<T,K> = {
            ...definition,
            handler: handlerFn
        };

        if(this.routes) this.routes.push(routeDeclaration)
        else this.routes = [routeDeclaration];

        return fn;
    }
}

/*
const testeSchema = {
    query: z.object({
        teste: z.string().optional()
    })
}

//Exemplo de uso:
class Teste extends ControllerBase {
    nome: string;
    constructor() {
        super();

        this.nome = "Nome salvo";
    }

    @route({
        method: "get",
        route: "/teste",
        schema: testeSchema
    })
    async teste(parsed: ParsedControllerSchema<typeof testeSchema>) {
        return {code: 200, data: this.nome};
    }
}

const teste = new Teste();

console.log("registrado:",teste.routes);

console.log(teste.teste({
    query: {
        teste: "teste"
    },
    body: undefined,
    params: undefined
}));

console.log(await (teste.routes[0] as any).handler({
    query: {
        teste: "teste"
    }
}, {
    locals: {
        tempoInicio: 0
    },
    status: () => ({
        json: (data: any) => console.log(data)
    })
},
    (e: any) => console.error(e))
);*/