import bcrypt from "bcryptjs";
import { APIError } from "../middlewares/errorHandler.js";
import { UsuarioRepository } from "../models/Usuario.js";
import { UsuarioSchemas } from "../schemas/usuarioSchemas.js";
import { getPaginatedResponseSchema, getResponseSchema } from "../utils/messages.js";
import ControllerBase, { ParsedSchema, route } from "./ControllerBase.js";

export default class UsuarioControler extends ControllerBase{
    public repo: UsuarioRepository;

    constructor(
        repo = UsuarioRepository.getInstance()
    ) {
        super("Usuario");

        this.repo = repo;
    }

    @route({
        method: "post",
        route: "/usuarios",
        schema: UsuarioSchemas.cadastrarUsuario,
        response: {code: 201, json: getResponseSchema(UsuarioSchemas.usuarioGET)}
    })
    async cadastrarUsuario({body}: ParsedSchema<typeof UsuarioSchemas.cadastrarUsuario>) {

        // Pesquisar usando collation, para ser case-insensitive
        const usernameDuplicado = await this.repo.obterUsuarioPorUsername(body.username, {ignoreCase: true});
        if(usernameDuplicado) {
            throw new APIError({message: "nome de usuário não disponível", path: "username"}, 400);
        }

        const emailDuplicado = await this.repo.obterUsuarioPorEmail(body.email);
        if(emailDuplicado) {
            throw new APIError({message: "email em uso", path: "email"}, 400);
        }

        const {senha, ...dados} = body;
        const senhaBcrypt = await bcrypt.hash(senha, await bcrypt.genSalt());
        const usuario = await this.repo.create({
            ...dados,
            senha: senhaBcrypt
        });

        return {code: 201, json: this.repo.omitUnselected(usuario)};
    }

    @route({
        method: "get",
        route: "/usuarios",
        schema: UsuarioSchemas.listarUsuarios,
        response: {code: 200, json: getPaginatedResponseSchema(UsuarioSchemas.usuarioGET)}
    })
    async listarUsuarios({query}: ParsedSchema<typeof UsuarioSchemas.listarUsuarios>) {
        const result = await this.repo.listar(query);

        return {code: 200, json: result};
    }
}