import { OpenAPIRegistry, OpenApiGeneratorV3 } from "@asteasolutions/zod-to-openapi";
import swaggerJSDoc from "swagger-jsdoc";
import SwaggerUI from "swagger-ui-express";
import express from "express";
import ControllerBase from "../controllers/ControllerBase.js";

export const getSwaggerOpts: () => SwaggerUI.SwaggerOptions = () => {
    return {
        // Ordenar as tags do swagger https://github.com/scottie1984/swagger-ui-express/issues/89
        swaggerOptions: {
            tagsSorter: (a: any, b: any) => { // Obs: esse código roda do lado browser. não pode usar nada tem que ser auto-contido
                const tagsOrder = ["codigo"];
                const result = tagsOrder.indexOf(a.toLowerCase()) - tagsOrder.indexOf(b.toLowerCase());

                if (result === 0) return a.localeCompare(b);
                else return result;
            },
            operationsSorter: (a: any, b: any) => { // Obs: esse código roda do lado browser. não pode usar nada tem que ser auto-contido
                const methodsOrder = ["get", "post", "put", "patch", "delete", "options", "trace"];
                const result = methodsOrder.indexOf(a.get("method")) - methodsOrder.indexOf(b.get("method"));

                if (result === 0) return a.get("path").localeCompare(b.get("path"));
                else return result;
            }
        }
    };
};

// para ficar o url certo sem precisar mudar
const getServersInCorrectOrder: () => [{ url: string }, { url: string }] = () => {
    const devUrl = { url: process.env.SWAGGER_DEV_URL || "http://localhost:3000" };
    const prodUrl = { url: process.env.SWAGGER_PROD_URL || "http://localhost:3000" };

    if (process.env.NODE_ENV === "production") return [prodUrl, devUrl];
    else return [devUrl, prodUrl];
};

function convertExpressRouteToOpenAPI(route: string) {
    return route.replace(/:(\w+)/g, '{$1}');
}

export const registerControllerPaths = (controller: ControllerBase, registry: OpenAPIRegistry) => {
    controller.routes.forEach(route => {
        const response = route.response || { code: 200 };
        registry.registerPath({
            method: route.method,
            path: convertExpressRouteToOpenAPI(route.route),
            description: route.description || `Rota ${route.route} do controller ${controller.name}`,
            summary: route.summary,
            tags: [controller.name],
            security: route.auth === undefined ? undefined : [{ "bearerAuth": [] }],
            request: {
                query: route.schema.query,
                params: route.schema.params,
                ...(route.schema.body ? {
                    body: {
                        content: {
                            "application/json": {
                                schema: route.schema.body
                            }
                        }
                    }
                } : {})
            },
            responses: {
                [response.code]: {
                    description: `Resposta de código ${response.code}`,
                    ...(response.json ? {
                        content: {
                            "application/json": {
                                schema: response.json
                            }
                        }
                    } : {})
                }
            },
        });
    });
};

export const rotasSwaggerDoc = async (registry: OpenAPIRegistry) => {
    // https://github.com/asteasolutions/zod-to-openapi
    // https://github.com/asteasolutions/zod-to-openapi/blob/master/example/index.ts
    const generator = new OpenApiGeneratorV3(registry.definitions);

    let ret = generator.generateDocument({
        openapi: "3.0.0",
        info: {
            title: "API AVALIADOR",
            version: "1.0-alpha",
            description: "API AVALIADOR\n\nÉ necessário autenticar com token JWT antes de utilizar a maioria das rotas, faça isso na rota /login com um email e senha válido\n\nMais informações em [https://gitlab.fslab.dev/erickweil/avaliador](https://gitlab.fslab.dev/erickweil/avaliador)",
            contact: {
                name: "Erick Leonardo Weil",
                email: "erick.weil@ifro.edu.br",
            },
        },
        servers: getServersInCorrectOrder()
    });

    const router = express.Router();
    router.use("/docs", SwaggerUI.serve, SwaggerUI.setup(swaggerJSDoc({
        swaggerDefinition: {
            ...ret
        },
        apis: []
    }), getSwaggerOpts()));

    return router;
};