import { Codigo, CodigoDoc, CodigoRepository, CodigoTeste, codigoStatus } from "../models/Codigo.js";
import { ExecAvaliar, ExecAvaliarLocal } from "../services/ExecAvaliar.js";
import { Problema, ProblemaDoc, ProblemaTeste } from "../models/Problema.js";
import { APIError } from "../middlewares/errorHandler.js";

export class CodigoInteractor {
    public execAvaliar: ExecAvaliar;
    public repo: CodigoRepository;

    constructor(
        execAvaliar: ExecAvaliar,
        codigoRepository = CodigoRepository.getInstance()
    ) {
        this.execAvaliar = execAvaliar;
        this.repo = codigoRepository;
    }

    static async getInstance() {
        let execAvaliar: ExecAvaliar;
        if(await ExecAvaliar.verificarDockerDisponivel()) {
            execAvaliar = new ExecAvaliar();
        } else {
            if(process.env.DEBUGLOG === "true") {
                console.log("===================================================");
                console.log("                   ATENÇÃO                         ");
                console.log("Docker não disponível, iniciando com execução local");
                console.log("===================================================");
            }
            execAvaliar = new ExecAvaliarLocal();
        }

        return new CodigoInteractor(execAvaliar);
    }

    private async iniciarTeste(codigo: CodigoDoc, teste: CodigoTeste) {
        codigo.testes.push(teste);
        await this.repo.patch(codigo);

        return codigo.testes[codigo.testes.length - 1];
    }

    private async resultadoTeste(codigo: CodigoDoc, resultadoTeste: CodigoTeste) {
        // alterou o objeto, vai funcionar confia, deixa o mongoose trabalhar
        await this.repo.patch(codigo);
    }

    private async atualizarStatusCodigo(codigo: CodigoDoc, status: typeof codigoStatus[number]) {
        codigo.status = status;
        if(status === "Avaliando") {
            codigo.data_recebido = new Date();
        } else {
            codigo.data_avaliado = new Date();
        }
        await this.repo.patch(codigo);
    }

    async enviarCodigo(problema: Problema, codigo: string, linguagem: string) {
        if(!problema.testes_privados) {
            throw new APIError("Problema não possui testes não é possível enviar código.", 400);
        }

        const codigoObj = await this.repo.create({
            codigo: codigo,
            problema: problema.sigla,
            linguagem: linguagem,
            status: "Enviado"
        });
        //console.log("Criado:", JSON.stringify(codigoObj,null,2));

        // Enviar código para avaliação
        // TODO: UTILIZAR FILAS COM RABBITMQ
        setTimeout(async () => {
            // Tem que ter try-catch para não travar a api
            // Já que essa função no setTimeout faz parte outra callStack o wrapException não pega
            try {
                //console.log("Avaliando...");
                await this.atualizarStatusCodigo(codigoObj, "Avaliando");

                const todosTestes = problema.testes.concat(problema.testes_privados || []);
                //console.log("Testes para avaliar:", todosTestes.map(t => t.num).join(", "));

                const resultado = await this.execAvaliar.avaliarTestesCodigo(codigoObj.codigo, codigoObj.linguagem, todosTestes, async (t) => {
                    //console.log("Iniciando teste", t.num);
                    return await this.iniciarTeste(codigoObj, t);
                }, async (t) => {
                    //console.log("Resultado teste", t.num, t.resultado);
                    await this.resultadoTeste(codigoObj, t);
                });

                //console.log("Resultado final:", resultado);
                await this.atualizarStatusCodigo(codigoObj, resultado.sucess ? "Sucesso" : "Erro");
            } catch(e) {
                // na pior da hipóteses, se der erro, marcar como erro para não ficar "Avaliando"
                console.error(e);
                
                // Obtém o objeto do banco para garantir que o erro não é o conteúdo do objeto manipulado
                try {
                    const codigoObjBanco = await this.repo.findById(codigoObj._id);
                    if(codigoObjBanco)
                    await this.atualizarStatusCodigo(codigoObjBanco, "Erro");
                } catch(e2) {
                    console.error("Erro ao atualizar o status para Erro!!!", e2);
                }
            }
        }, 0);

        let ret = this.repo.omitUnselected(codigoObj);
        //console.log("Criado:", JSON.stringify(ret,null,2));
        return ret;
    }
}