import { ErrorRequestHandler, NextFunction, Request, Response } from "express";
import { ZodError } from "zod";
import { sendError } from "../utils/messages.js";

export class APIError extends Error {
    public code: number;
    public errors: { message: string, path?: string }[];

    constructor(errors: string|{ message: string, path?: string }|{ message: string, path?: string }[], 
        code = 400, 
        options = undefined) {
        super(errors ? ""+errors : "Erro com código " + code, options);
        this.code = code;

        if(Array.isArray(errors)) { 
            this.errors = errors;
        } else if(typeof errors === "object" && errors.message !== undefined) {
            this.errors = [errors];
        } else {
            this.errors = [{message: ""+errors}];
        }
    }
}

/**
    https://expressjs.com/en/guide/using-middleware.html#middleware.error-handling
    Error-handling middleware always takes four arguments. You must provide four arguments to identify it as an error-handling middleware function. 
    Even if you don’t need to use the next object, you must specify it to maintain the signature. 
    Otherwise, the next object will be interpreted as regular middleware and will fail to handle errors.

    https://expressjs.com/en/guide/error-handling.html
    Errors that occur in synchronous code inside route handlers and middleware require no extra work. 
    If synchronous code throws an error, then Express will catch and process it.

    Starting with Express 5, route handlers and middleware that return a Promise will call next(value) automatically when they reject or throw an error.

    You define error-handling middleware last, after other app.use() and routes calls;
*/
export const errorHandler: ErrorRequestHandler = (err: Error, req, res, next) => {
    if (err instanceof APIError) {
        // Quando é um erro que eu mesmo criei, com código e mensagens específicas
        sendError(res, err.code, err.errors);
    } else if (err instanceof ZodError) {
        // Erro de validação do Zod
        let errors = [];
        for(const issue of err.issues) {
            errors.push({
                message: issue.message,
                path: issue.path.join(".")
            });
        }

        sendError(res, 422, errors);
    } else if (err.name === "ValidationError") {
        // Erros de validação do mongoose
        let errors: {message: string, path?: string}[] = [{ message: "Erros no model do Mongoose" }];

        if("errors" in err && typeof err.errors === "object" && err.errors !== null && !Array.isArray(err.errors)) {
            Object.entries(err.errors).forEach(([key, value]: [string, unknown]) => {
                errors.push({
                    message: ""+(typeof value === "object" && value !== null && "message" in value ? value.message : ""+value),
                    path: key
                });
            });
        }

        sendError(res, 422, errors);
    } else {
        // Erro desconhecido
        sendError(res, 500, err.message || "" + err);
    }
};