import mongoose from "mongoose";
import { MongooseRepository, RepoDoc } from "../utils/mongooseRepository.js";
import { defaultSchemaOptions, SchemaBase } from "../config/mongooseConfigs.js";
import paginate from "mongoose-paginate-v2";

export const codigoTesteResultado = ["Executando", "Sucesso", "Erro", "Timeout", "Diferente", "Formatacao"] as const;
export const codigoStatus = ["Enviado", "Avaliando", "Sucesso", "Erro"] as const;

export interface CodigoTeste extends SchemaBase {
    num: number;
    saida?: string;
    resultado: typeof codigoTesteResultado[number];
    argumentos?: string[];
    distancia?: number;
    tempo?: number;
    data_processado?: Date;
}

export interface Codigo extends SchemaBase {
    codigo: string;
    problema: string;
    linguagem: string;
    status: typeof codigoStatus[number];
    data_envio: Date;
    testes: CodigoTeste[];
    usuario?: mongoose.Types.ObjectId;
    data_recebido?: Date;
    data_avaliado?: Date;
}

const CodigoSchema = new mongoose.Schema<Codigo>({
    codigo: {
        type: String,
        required: true
    },
    problema: {
        type: String,
        required: true
    },
    linguagem: {
        type: String,
        required: true
    },
    status: {
        type: String,
        enum: codigoStatus,
        required: true
    },
    data_envio: {
        type: Date,
        default: Date.now
    },
    testes: [{
        num: {
            type: Number,
            required: true
        },
        saida: {
            type: String,
            required: false
        },
        resultado: {
            type: String,
            enum: codigoTesteResultado,
            required: true
        },
        argumentos: {
            type: [String],
            default: []
        },
        distancia: {
            type: Number,
            required: false
        },
        tempo: {
            type: Number,
            required: false
        },
        data_processado: {
            type: Date,
            required: false
        }
    }],
    usuario: {
        type: mongoose.Schema.Types.ObjectId,
        ref: "usuarios",
        required: false
    },
    data_recebido: {
        type: Date,
        required: false
    },
    data_avaliado: {
        type: Date,
        required: false
    }
}, defaultSchemaOptions());

CodigoSchema.plugin(paginate);

export const getCodigoModel = () => mongoose.model<Codigo, mongoose.PaginateModel<Codigo>>("codigos", CodigoSchema);

export class CodigoRepository extends MongooseRepository<Codigo, mongoose.PaginateModel<Codigo>> {

    static instance: CodigoRepository;

    constructor() {
        super(getCodigoModel());
    }

    static getInstance() {
        if (!this.instance) this.instance = new CodigoRepository();
        return this.instance;
    }

    async listar({pagina, limite}: {pagina?: number, limite?: number}) {
        let opcoes: mongoose.PaginateOptions = {};
        if(pagina) opcoes.page = pagina;
        if(limite) opcoes.limit = limite;
        opcoes.sort = { _id: -1 }; // Mais recentes primeiro

        return await this.model.paginate({},opcoes);
    }
}
export type CodigoDoc = RepoDoc<Codigo>;