import mongoose from "mongoose";
import { MongooseRepository, RepoDoc } from "../utils/mongooseRepository.js";
import { defaultSchemaOptions, SchemaBase, } from "../config/mongooseConfigs.js";
import paginate from "mongoose-paginate-v2";

export interface ProblemaTeste extends SchemaBase {
    num: number;

    // Argumentos que serão passados para o código
    argumentos: string[];

    // Resultado esperado
    saida: string;

    // Indica se a saída esperada é um regex ou não
    regex: boolean;
}

export interface Problema extends SchemaBase {
    // Identificador único legível, utilizado como slug no url
    sigla: string;

    titulo: string;

    // texto do problema
    descricao: string;

    // HTML sanitizado do texto do problema
    descricao_html: string;

    testes: ProblemaTeste[];

    testes_privados?: ProblemaTeste[];

    // testes.length + testes_privados.length
    testes_quantidade: number;
}

export const ProblemaTesteSchema = new mongoose.Schema<ProblemaTeste>({
    num: { 
        type: Number, 
        required: true 
    },
    argumentos: { 
        type: [String], 
        default: [] 
    },
    saida: { 
        type: String, 
        required: true 
    },
    regex: { 
        type: Boolean, 
        default: false 
    }
}, defaultSchemaOptions());

export const ProblemaSchema = new mongoose.Schema<Problema>({
    sigla: { 
        type: String, 
        required: true, 
        unique: true 
    },
    titulo: { 
        type: String, 
        required: true 
    },
    descricao: { 
        type: String, 
        required: true 
    },
    descricao_html: { 
        type: String, 
        required: true 
    },
    testes: { 
        type: [ProblemaTesteSchema], 
        default: [] 
    },
    testes_privados: { 
        type: [ProblemaTesteSchema], 
        default: [], 
        select: false 
    },
    testes_quantidade: { 
        type: Number, 
        default: 0 
    }
}, defaultSchemaOptions());

ProblemaSchema.plugin(paginate);

export const getProblemaModel = () => mongoose.model<Problema, mongoose.PaginateModel<Problema>>("problemas", ProblemaSchema);

export class ProblemaRepository extends MongooseRepository<Problema, mongoose.PaginateModel<Problema>> {

    static instance: ProblemaRepository;

    constructor() {
        super(getProblemaModel());
    }

    static getInstance() {
        if (!this.instance) this.instance = new ProblemaRepository();
        return this.instance;
    }

    async listar({pagina, limite}: {pagina?: number, limite?: number}) {
        let opcoes: mongoose.PaginateOptions = {};
        if(pagina) opcoes.page = pagina;
        if(limite) opcoes.limit = limite;
        opcoes.sort = { _id: 1 }; // Mais antigos primeiro
        opcoes.projection = { testes: { $slice: 3 } }; // Limita a quantidade de testes retornados

        return await this.model.paginate({},opcoes);
    }

    async obterPorSigla(sigla: string, testes_privados = false) {
        let query = this.model.findOne({ 
            sigla: sigla.toLowerCase() 
        });
        if(testes_privados) query = query.select("+testes_privados");
        return await query.exec();
    }
}
export type ProblemaDoc = RepoDoc<Problema>;