import mongoose from "mongoose";
import { defaultSchemaOptions, SchemaBase } from "../config/mongooseConfigs.js";
import { MongooseRepository, RepoDoc } from "../utils/mongooseRepository.js";
import paginate from "mongoose-paginate-v2";
import { APIError } from "../middlewares/errorHandler.js";

export interface Usuario extends SchemaBase {
    nome: string,
    username: string,

    email: string,
    email_confirmado: boolean,
    codigo_confirmacao?: string,

    senha?: string,
    codigo_recuperacao?: string,

    banido_ate?: Date,

    created_at: Date,
}

const UsuarioSchema = new mongoose.Schema<Usuario>({
    nome: {
        type: String, 
        required: true
    },
    username: {
        type: String, 
        required: true
    },
    email: {
        type: String, 
        required: true, 
        unique: true,
        select: false
    },
    email_confirmado: {
        type: Boolean, 
        default: false
    },
    codigo_confirmacao: {
        type: String,
        required: false,
        select: false
    },
    senha: {
        type: String,
        required: false,
        select: false
    },
    codigo_recuperacao: {
        type: String,
        required: false,
        select: false
    },
    banido_ate: {
        type: Date,
        required: false
    },
    created_at: {
        type: Date, 
        default: Date.now
    }
}, defaultSchemaOptions());

// https://www.mongodb.com/docs/manual/core/index-case-insensitive/
// https://stackoverflow.com/questions/33736192/mongo-unique-index-case-insensitive
// Este índice garante que username será único mesmo que haja diferença entre maiúsculas e minúsculas
// Porém não é para usar esse índice na busca, na verdade geralmente a busca vai ser case-sensitive normal
UsuarioSchema.index(
    { username: 1 },
    {
      collation: {locale: "pt", strength: 2}, // ignora maiúsculas e minúsculas ao comparar
      unique: true
    }
);

UsuarioSchema.plugin(paginate);
export const getUsuarioModel = () => mongoose.model<Usuario, mongoose.PaginateModel<Usuario>>("usuarios", UsuarioSchema);

export class UsuarioRepository extends MongooseRepository<Usuario, mongoose.PaginateModel<Usuario>> {

    static instance: UsuarioRepository;
    
    constructor() {
        super(getUsuarioModel());
    }

    static getInstance() {
        if (!this.instance) this.instance = new UsuarioRepository();
        return this.instance;
    }

    async obterUsuarioPorUsername(username: string, options: {ignoreCase?: boolean, selectSenha?: boolean} = {}) {
        let query = this.model.findOne({
            username: username
        });
        if(options.ignoreCase) query = query.collation({locale: "pt", strength: 2});
        if(options.selectSenha) query = query.select("+senha");

        return await query.exec();
    }

    async obterUsuarioPorEmail(email: string, options: {selectSenha?: boolean} = {}) {
        let query = this.model.findOne({
            email: email
        });
        if(options.selectSenha) query = query.select("+senha");

        return await query.exec();
    }

    async listar({pagina, limite, username}: {pagina?: number, limite?: number, username?: string}) {
        let opcoes: mongoose.PaginateOptions = {};
        if(pagina) opcoes.page = pagina;
        if(limite) opcoes.limit = limite;
        opcoes.sort = { _id: -1 }; // Mais recentes primeiro

        let filtros = {};
        if(username) filtros = {username: new RegExp(`^${username}`)};

        return await this.model.paginate(filtros, opcoes);
    }
}
export type UsuarioDoc = RepoDoc<Usuario>;