import logRoutes from "../middlewares/LogRoutesMiddleware.js";
import { Router } from "express";
import { OpenAPIRegistry } from "@asteasolutions/zod-to-openapi";
import { registerControllerPaths, rotasSwaggerDoc } from "../docs/head.js";
import { APIError, errorHandler } from "../middlewares/errorHandler.js";
import ControllerBase from "../controllers/ControllerBase.js";
import CodigoControler from "../controllers/CodigoController.js";
import ProblemaController from "../controllers/ProblemaController.js";
import UsuarioControler from "../controllers/UsuarioController.js";
import AuthMiddleware from "../middlewares/authMiddleware.js";
import AuthController from "../controllers/AuthController.js";

export const getRoutes = (controller: ControllerBase, registry: OpenAPIRegistry) => {
    const router = Router();
            
    // Documentação
    registerControllerPaths(controller, registry);

    // Rotas
    controller.routes.forEach(route => {
        if(!route.auth)
            router[route.method](route.route, route.handler);
        else {
            router[route.method](route.route, AuthMiddleware(route.auth), route.handler);
        }
    });
    
    return router;
}

const routes = async (app: Router) => {

    if (process.env.DEBUGLOG === "true") {
        app.use(logRoutes);
    }

    app.get("/", (req, res) => {
        res.status(200).redirect("docs"); // redirecionando para documentação
    });

    const registry = new OpenAPIRegistry();
    const bearerAuth = registry.registerComponent("securitySchemes", "bearerAuth", {
        type: "http",
        scheme: "bearer",
        bearerFormat: "JWT",
    });

    app.use(
        getRoutes(await CodigoControler.getInstance(), registry),
        getRoutes(new ProblemaController(), registry),
        getRoutes(new UsuarioControler(), registry),
        getRoutes(new AuthController(), registry)
    );

    app.use(await rotasSwaggerDoc(registry));

    app.use((req, res, next) => {
        throw new APIError("Rota não encontrada", 404);
    });

    // Middleware de tratamento de erros, sempre por último
    app.use(errorHandler);
};

export default routes;
