// A ideia é que irá executar a avaliação de um Código para cada um dos casos de teste
import { randomUUID } from "crypto";
import { unlink, mkdir, writeFile, rename, readFile } from "fs/promises";
import { bashExec } from "../utils/bashExec.js";
import { semelhancaTexto } from "../utils/compararTexto.js";
import anyAscii from "any-ascii";
import { ProblemaTeste } from "../models/Problema.js";
import { CodigoTeste, codigoTesteResultado } from "../models/Codigo.js";

type LinguagemConfig = {
    dockerfile: string,
    imagem: string,
    extensao: string,
    comando: string[],
    enviarEntrada: boolean
};
export const configuracaoLinguagens = new Map<string, LinguagemConfig>([
    ["golang", {
        "dockerfile": "golang.Dockerfile",
        "imagem": "avaliar-golang",
        "extensao": "go",
        "comando": ["go","run"],
        "enviarEntrada": false
    }],
    ["javascript", {
        "dockerfile": "javascript.Dockerfile",
        "imagem": "avaliar-javascript",
        "extensao": "mjs",
        "comando": ["node"],
        "enviarEntrada": false
    }],
    ["java", {
        "dockerfile": "java.Dockerfile",
        "imagem": "avaliar-java",
        "extensao": "java",
        "comando": ["java"],
        "enviarEntrada": false
    }],
    ["python", {
        "dockerfile": "python.Dockerfile",
        "imagem": "avaliar-python",
        "extensao": "py",
        "comando": ["python"],
        "enviarEntrada": false
    }],
    ["portugol", {
        "dockerfile": "portugol.Dockerfile",
        "imagem": "avaliar-portugol",
        "extensao": "por",
        "comando": ["node","/app/terminal.js","--tempo","0","--programa"],
        "enviarEntrada": true
    }],
]);
export const LINGUAGENS = ["golang","javascript","java","python","portugol"] as const;

type ResultadoExecAvaliarTeste = {
    saida: string;
    tempo: number;
    resultado: "Sucesso" | "Erro" | "Timeout" | "Diferente" | "Formatacao";
    distancia: number | undefined;
};

export class ExecAvaliar {
    protected codeDir: string;
    protected execTimeout: number;

    public constructor(
        codeDir = process.env.CODE_PATH,
        execTimeout: string | number | undefined = process.env.EXEC_TIMEOUT
    ) {
        this.codeDir = codeDir || "./code";
        this.execTimeout = execTimeout !== undefined ? typeof execTimeout === "number" ? execTimeout : parseInt(execTimeout) : 10000;
    }

    public static async verificarDockerDisponivel() {
        // command -v docker && echo "OK" || echo "NAO"
        const result = await bashExec("sh", ["-c","command -v docker"], undefined, 10000);
        return result.sucess;
    }

    public async criarImagens() {        
        await mkdir(this.codeDir, { recursive: true });

        for(const [linguagem,config] of configuracaoLinguagens.entries()) {
            console.log("Criando imagem para linguagem", linguagem);
            await bashExec("docker",[
                "build",
                "-f", `./docker/${config.dockerfile}`,
                "-t", config.imagem,
                "."
            ],
            undefined,
            10 * 60000 // 10 minutos de timeout para build de cada imagem é suficiente?
            );
        }
    }

    protected async iniciarContainer(imagem: string): Promise<string> {
        const uuid = randomUUID();
        // https://medium.com/@maheshwar.ramkrushna/best-practices-for-secure-docker-containerization-non-root-user-read-only-volumes-and-resource-d34ed09b1bd3
        const result = await bashExec("docker", [
            "run",
            "-d",
            "--rm",
            "--name", uuid,
            "--read-only",
            "--network","none",
            "--memory","512M",
            "--memory-reservation","512M",
            "--cpus","2",
            "--pids-limit","128",
            "-v",`./${this.codeDir}:/code:ro`,
            imagem
        ],
        undefined,
        // timeout
        this.execTimeout,
        // killSignal
        "SIGKILL"
        );

        if(!result.sucess) {
            throw new Error("Não foi possível iniciar o container:"+result.output);
        }

        return uuid;
    }

    protected async pararContainer(uuid: string): Promise<void> {
        await bashExec("docker",["rm","-f",uuid]);
    }

    protected async executarCodigo(uuid: string, arquivoCodigo: string, entrada: string | undefined, argumentos: string[], cmd: string[]) {
        return await bashExec("docker",[
            "exec",
            "-i",
            uuid,
            ...cmd,
            "/code/"+arquivoCodigo,
            ...argumentos
        ],
        entrada,
        // timeout
        this.execTimeout,
        // enviar ctrl+c se der timeout
        "SIGINT");
    }

    protected async avaliarTeste(uuid: string, filename: string, config: LinguagemConfig, teste: ProblemaTeste): Promise<ResultadoExecAvaliarTeste> {
        let argumentos: string[] = [];
        let entrada: string | undefined = undefined;
        if(config.enviarEntrada) {
            entrada = teste.argumentos.join("\n");
        } else {
            argumentos = teste.argumentos;
        }

        const resultado = await this.executarCodigo(uuid, filename, entrada, argumentos, config.comando);
        let resultadoFeedback: typeof codigoTesteResultado[number] = "Sucesso";
        let distancia: number | undefined = undefined;

        if(!resultado.sucess) {
            resultadoFeedback = resultado.duration >= this.execTimeout ? "Timeout" : "Erro";
            if(process.env.DEBUGLOG === "true" && resultado.output) {
                console.error(resultado.output);
            }
        } else {
            distancia = semelhancaTexto(resultado.output, teste.saida, teste.regex);

            if(distancia != 0) {
                resultadoFeedback = "Diferente";
                // Verificar se com o texto sem espaços e acentos é igual
                if(!teste.regex) {
                    const outputSemEspacos = anyAscii(resultado.output.replace(/\s/g, "").toLowerCase());
                    const saidaSemEspacos = anyAscii(teste.saida.replace(/\s/g, "").toLowerCase());
                    if(outputSemEspacos === saidaSemEspacos) {
                        resultadoFeedback = "Formatacao";
                    }
                }
            }
        }

        return {
            saida: resultado.output,
            tempo: resultado.duration,
            resultado: resultadoFeedback,
            distancia: distancia
        };
    }

    public async avaliarTestesCodigo(codigo: string, linguagem: string, testes: ProblemaTeste[],
        onIniciarTeste: (teste: CodigoTeste) => Promise<CodigoTeste>,
        onResultadoTeste?: (teste: CodigoTeste) => Promise<void>) {
        if(!testes || testes.length === 0) {
            throw new Error("Problema sem testes");
        }

        const config = configuracaoLinguagens.get(linguagem);
        // type guard linguagem must be in configuracaoLinguagens
        if(!config) {
            throw new Error("Linguagem inválida");
         }

        let uuid: string | undefined = undefined;
        let filename: string | undefined = undefined;
        let filepath: string | undefined = undefined;
        let resultadoTestes: CodigoTeste[] = [];
        try {
            uuid = await this.iniciarContainer(config.imagem);
            filename = uuid+"."+config.extensao;
            filepath = this.codeDir+"/"+filename;
            await writeFile(filepath, codigo);

            for(let teste of testes) {
                const codigoTeste = await onIniciarTeste({
                    num: teste.num,
                    argumentos: teste.argumentos,
                    resultado: "Executando"
                });
                resultadoTestes.push(codigoTeste);

                const resultado = await this.avaliarTeste(uuid, filename, config, teste);
                codigoTeste.saida = resultado.saida;
                codigoTeste.tempo = resultado.tempo;
                codigoTeste.resultado = resultado.resultado;
                codigoTeste.distancia = resultado.distancia;
                codigoTeste.data_processado = new Date();
                if(onResultadoTeste) {
                    await onResultadoTeste(codigoTeste);
                }
                
                // Só executar os outros testes se der sucesso
                if(resultado.distancia !== 0) {
                    return {sucess: false, testes: resultadoTestes};
                }
            }

            return {sucess: true, testes: resultadoTestes};
        } catch(err: any) {
            console.error(err);
            return {sucess: false, testes: resultadoTestes, output: err.message || err};
        } finally {
            // parar container
            if(uuid) {
                await this.pararContainer(uuid);
            }
            // deletar arquivo
            if(filepath)
            await unlink(filepath);
        }
    }
}

export class ExecAvaliarLocal extends ExecAvaliar {
    public async criarImagens(): Promise<void> {
        // Não precisa criar imagens
    }

    protected async iniciarContainer(imagem: string): Promise<string> {
        if(imagem === "avaliar-portugol") {
            throw new Error("Não é possível executar Portugol localmente");
        }
        return imagem;
    }

    protected async pararContainer(uuid: string): Promise<void> {
        // Não precisa parar container
    }

    protected async executarCodigo(uuid: string, arquivoCodigo: string, entrada: string | undefined, argumentos: string[], cmd: string[]) {
        const [ first, ...rest ] = cmd;
        const result = await bashExec(first,[
            ...rest,
            this.codeDir+"/"+arquivoCodigo,
            ...argumentos
        ],
        entrada,
        // timeout
        this.execTimeout,
        // enviar KILL se der timeout
        "SIGKILL");

        return result;
    }
}