import { Request, Response, NextFunction } from "express";
import mongoose from "mongoose";
import { ZodTypeAny, z } from "zod";
import { RepoDoc } from "./mongooseRepository.js";
import { SchemaBase } from "../config/mongooseConfigs.js";

/**
 * Envia uma resposta de erro com o código e a mensagem especificada
 * de acordo com o padrão de envio de respostas da API
 * 
 * @example
 * //Exemplos de diferentes formas aceitáveis de usar:
 * sendError(res, 400, "Mensagem de erro")
 * sendError(res, 400, [{message:"msg A"}, {message:"msg B"}])
 * sendError(res, 400, {message:"Campo Obrigatório", field:"senha"})
 */
export const getErrorResponseSchema = () => z.object({
    data: z.null(),
    errors: z.array(z.object({
        message: z.string(),
        path: z.string().optional()
    }))
});

export const sendError = (
    res: Response, 
    code: number, 
    errors: string|{message?: string}|{message?: string}[] = []) => {
    try {
        // Detecta diferentes formas de usar:
        let _errors: {message?: string | undefined}[];
        if(Array.isArray(errors)) { 
            // Se for um array de erros --> sendError(res, 400, [{message:"A"},{message:"B"}])
            _errors = errors;
        } else if(typeof errors === "object" && errors.message !== undefined) {
            // Se for um objeto com a propriedade message --> sendError(res, 400, {message:"A"})
            _errors = [errors];
        } else {
            // Se for uma string ou qualquer outro tipo --> sendError(res, 400, "A")
            _errors = [{message: ""+errors}];
        }

        if(code === 500) {
            console.error(errors);
        }
        return res.status(code).json({
            data: null,
            errors: _errors,
        });
    } finally {
        if (process.env.DEBUGLOG === "true") {
            const millis = Math.floor(performance.now() - res.locals.tempoInicio!);
            console.log(millis + " ms");
        }
    }
};

export const getPaginatedResponseSchema = <T extends ZodTypeAny>(dataSchema: T) => z.object({
    data: z.array(dataSchema),
    meta: z.object({
        limite: z.number(),
        pagina: z.number(),
        resultados: z.number()
    }),
    errors: z.null()
});

export const getResponseSchema = <T extends ZodTypeAny>(dataSchema: T) => z.object({
    data: dataSchema,
    errors: z.null()
});

export type ControllerResp<T = object> = 
      mongoose.PaginateResult<RepoDoc<T>> 
    | RepoDoc<T> 
    | RepoDoc<T>[]
    | T 
    | T[];
/**
 * Envia uma resposta com o código e a mensagem especificada
 * de acordo com o padrão de envio de respostas da API
 * 
 * @example
 * sendResponse(res, 200, usuario);
 */
export const sendResponse = (res: Response, code: number, 
    resp?: ControllerResp) => {
    try {
        let dataToSend: unknown;
        let meta: Record<string, unknown> | undefined = undefined;
        if(!Array.isArray(resp) && typeof resp === "object" && "docs" in resp && "limit" in resp && "page" in resp && "totalDocs" in resp) {
            dataToSend = resp.docs;
            meta = {
                limite: resp.limit,
                pagina: resp.page,
                resultados: resp.totalDocs
            };
        } else {
            dataToSend = resp;
        }

        if(meta) {
            return res.status(code).json({
                data: dataToSend,
                meta: meta,
                errors: null
            });
        } else {
            return res.status(code).json({
                data: dataToSend,
                errors: null
            });
        }
    } finally {
        if (process.env.DEBUGLOG === "true") {
            const millis = Math.floor(performance.now() - res.locals.tempoInicio!);
            console.log(millis + " ms");
        }
    }
};