// Classe a ser extendida, que já possui métodos básicos padrão de acesso ao banco

import mongoose from "mongoose";
import { APIError } from "../middlewares/errorHandler.js";

export class RepositoryError extends APIError {
    constructor(errors: string|{ message: string, path?: string }|{ message: string, path?: string }[]) {
        super(errors, 422);
    }
}

type OptionalId = {
    _id?: mongoose.Types.ObjectId;
};


/*export type RepoModel<T, K> = mongoose.Model<T, // raw doc type
        Record<string, any>, // query helpers
        any, // instance methods
        // eslint-disable-next-line @typescript-eslint/ban-types
        {}> & K;
*/
//export type RepoDoc<T> = mongoose.Document<unknown, Record<string, any>, T> & T;
export type RepoDoc<T> = mongoose.Document<unknown, Record<string, any>, T> & T & Required<{
    _id: mongoose.Types.ObjectId;
    __v: number;
}>;

export type ModelInterface<M> = M extends mongoose.Model<infer T> ? T : never;

const isObject = (value: unknown) => {
    if(typeof value !== "object" || Array.isArray(value) || value === null) {
        return false;
    } else return true;
};

export const mongooseIdEquals = (id1: unknown, id2: unknown) => {
    if(!id1 || !id2) return false;

    let isId1 = id1 instanceof mongoose.Types.ObjectId;
    let isId2 = id2 instanceof mongoose.Types.ObjectId;

    if(isId1 && isId2) {
        return (id1 as mongoose.Types.ObjectId).equals(id2 as mongoose.Types.ObjectId);
    }

    if(isId1) id1 = id1.toString();
    if(isId2) id2 = id2.toString();
    return id1 === id2;
};

export function patch(doc: Record<string, any>, data: Record<string, any>, options = {
    nullIsUndefined: false, // Se deve considerar campos que são null como undefined
    unsetUndefined: false, // Se quando for undefined deve remover o campo
}) {
    for(let key in data) {
        let value = data[key];

        if(options.nullIsUndefined && value === null) 
            value = undefined;

        if(!options.unsetUndefined && value === undefined) {
            continue;
        }

        doc[key] = value;
    }
}

/**
 * Não é typescript, mas é possível adicionar tipos com JSDoc!!
 * https://github.com/google/closure-compiler/wiki/Annotating-JavaScript-for-the-Closure-Compiler
 * https://stackoverflow.com/questions/8407622/set-type-for-function-parameters
 * 
 * https://stackoverflow.com/questions/38915195/jsdoc-mongoose-how-to-document-mongoose-models
 * https://www.typescriptlang.org/docs/handbook/utility-types.html
 */

//type ModelType<T> = T extends mongoose.Model<infer U> ? U : never;
//export type ModelDoc<T> = T extends mongoose.Model<infer U> ? InstanceType<mongoose.Model<U>> : never;
//export type ModelDoc<K> = mongoose.Document<mongoose.Types.ObjectId, {}, K> & K;


/**
 * Classe base para repositórios que utilizam o mongoose
 * T -> Interface do documento
 * M -> Model<T>
 */
export class MongooseRepository<T extends OptionalId, M extends mongoose.Model<T>> {
    public model: M;

    constructor(model: M) {
        this.model = model;
    }

    // https://www.typescriptlang.org/docs/handbook/advanced-types.html
    // type guard against T and RepoDoc<T>
    public isRepoDoc(doc: RepoDoc<T> | Partial<T>): doc is RepoDoc<T> {
        return (doc as RepoDoc<T>).db instanceof mongoose.Connection;
    }

    public omitUnselected(doc: RepoDoc<T>): Partial<T> {
        let ret = doc.toObject();

        this.model.schema.eachPath((path, schemaType) => {
            if(schemaType.options.select === false) {
                delete (ret as any)[path];
            }
        });

        return ret;
    }

    /**
     * Busca um documento pelo _id
     * 
     * retorna o documento encontrado, ou null se não encontrar
     */
    async findById(_id: string|mongoose.Types.ObjectId): Promise<RepoDoc<T> | null> {
        let query = this.model.findById(_id);
        return await query.exec();
    }

    /**
     * Criar um documento no banco de dados
     * 
     * retorna o documento criado ou um erro caso não consiga criar
     */
    async create(data: RepoDoc<T>|Partial<T>): Promise<RepoDoc<T>> {
        let _data: RepoDoc<T>;
        if(!(this.isRepoDoc(data))) {
            _data = new this.model(data);
        } else {
            _data = data;
        }

        return await _data.save();
    }

    /**
     * Atualiza um documento no banco de dados
     * 
     * retorna o documento atualizado ou um erro caso não consiga atualizar
     */
    async patch(doc: RepoDoc<T>, data?: Partial<T>, options = {
        nullIsUndefined: false, // Se deve considerar campos que são null como undefined
        unsetUndefined: false, // Se quando for undefined deve remover o campo
    }): Promise<RepoDoc<T>> {
        if(data) {
            patch(doc, data, options);
        }
        
        return await doc.save();
    }

    /**
     * Deleta um documento do banco de dados, baseado no _id
     * 
     * retorna true se deletou, false se não deletou (não encontrou)
     */
    async delete(doc: RepoDoc<T> | { _id: mongoose.Types.ObjectId }): Promise<boolean> {
        let result = await this.model.deleteOne({ _id: doc._id });
        return result.deletedCount > 0;
    }

    // Usado pelo seed para deletar todos os documentos
    async deleteEverything() {
        await this.model.deleteMany({});
    }
}