import { nanoid } from "nanoid";
import slug from "slug";

/* 
https://github.com/Trott/slug
slug: Slugifies strings, even when they contain Unicode.
Make strings URL-safe.
- Respects RFC 3986
- No dependencies
- Works in browser (window.slug) and AMD/CommonJS-flavoured module loaders

 
https://github.com/ai/nanoid/?tab=readme-ov-file
nanoid: A tiny, secure, URL-friendly, unique string ID generator for JavaScript.

“An amazing level of senseless perfectionism, which is simply impossible not to respect.”

- Small. 116 bytes (minified and brotlied). No dependencies. Size Limit controls the size.
- Safe. It uses hardware random generator. Can be used in clusters.
- Short IDs. It uses a larger alphabet than UUID (A-Za-z0-9_-). So ID size was reduced from 36 to 21 symbols.
- Portable. Nano ID was ported to over 20 programming languages.
*/
// Por fazer toLowerCase estou aumentando bastante a chance de colisões, mas tudo bem
const prefixLowercaseNanoId = (str: string, size?: number): string => {
    let nano = nanoid(size).toLowerCase();

    return nano+"-"+str;
};

const truncar = (str: string, maximo: number): string => {
    if(str.length > maximo) {
        return str.substring(0, maximo);
    } else {
        return str;
    }
};

/**
 * Gera um slug único a partir de um texto. Só se houver colisão prefixa com um nano id.
 */
export const gerarSlugUnico = async (
    texto: string, 
    checarColisao: (s: string) => Promise<boolean>,
    opcoes?: {
        minimo?: number,
        maximo?: number
    }
) => {
    let minimo = opcoes?.minimo || 8;
    let maximo = opcoes?.maximo || 160; // + ou menos o tamanho da barra de endereço em monitor 1080p
    
    let slugUnico = slug(texto, {
        replacement: "-",    
        lower: true,
        trim: true,
        fallback: true
    });
    
    if(slugUnico.length < minimo) {
        // Se o slug for muito curto, vamos adicionar um nano id
        // 10 - 7 = 3,   10 - 0 = 10
        // no mínimo 3 caracteres aleatórios, no máximo 10
        slugUnico = prefixLowercaseNanoId(slugUnico, (minimo - slugUnico.length)+2);
    }

    // Se o slug for muito longo, vamos truncar
    slugUnico = truncar(slugUnico, maximo);

    // Se houver colisão, vamos prefixar um nano id de 21 caracteres
    if(await checarColisao(slugUnico)) {
        // nanoid-slug
        slugUnico = prefixLowercaseNanoId(slugUnico);
        slugUnico = truncar(slugUnico, maximo);

        // Se houver colisão DENOVO, vamos prefixar um nano id de 21 caracteres DENOVO
        // sim vai ficar nanoid1-nanoid2-slug
        if(await checarColisao(slugUnico)) {
            slugUnico = prefixLowercaseNanoId(slugUnico);
            slugUnico = truncar(slugUnico, maximo);

            // A próxima colisão (improvável que aconteça) irá lançar erro de índice único ao salvar no banco de dados
        }
    }

    return slugUnico;
};