import { extendZodWithOpenApi } from "@asteasolutions/zod-to-openapi";
import mongoose from "mongoose";
import { ZodIssue, ZodIssueCode, ZodIssueOptionalMessage, z } from "zod";

export const myZ = {
    toMongooseId: () => z.string()
        .refine((id) => mongoose.Types.ObjectId.isValid(id), "id em formato inválido")
        .transform((id) => new mongoose.Types.ObjectId(id)),

    strToInt: () => z.string()
        .regex(/^-?[0-9]+$/, "número inteiro inválido")
        .transform((str) => parseInt(str, 10)),

    toDateTime: () => z.string()
        .datetime({ message: "Formato de data inválido", local: true})
        .transform((str) => new Date(str)),

    toUTCDate: () => z.string()
        .date("Formato de data inválido")
        .transform((str) => new Date(str)),

};

export const commonSchemas = {
    getPorId: () => ({
        params: z.object({
            id: myZ.toMongooseId()
        }).strict()
    }),

    listagemPaginada: () => (
        z.object({
            pagina: myZ.strToInt().pipe(z.number().min(1).max(9999)).optional(),
            limite: myZ.strToInt().pipe(z.number().min(0).max(100)).optional()
        }).strict()
    )
};

function traduzirNomeTipo(tipo: string): string {
    switch (tipo) {
        case "string": return "texto";
        case "nan": return "número inválido";
        case "number": return "número";
        case "integer": return "inteiro";
        case "float": return "número com casas decimais";
        case "boolean": return "verdadeiro ou falso";
        case "date": return "data";
        case "bigint": return "número grande";
        case "symbol": return "símbolo";
        case "function": return "função";
        case "undefined": return "nada";
        case "null": return "nulo";
        case "array": return "array";
        case "object": return "objeto";
        case "unknown": return "desconhecido";
        case "promise": return "promessa";
        case "void": return "vazio";
        case "never": return "nunca";
        case "map": return "mapa";
        case "set": return "conjunto";
        default:
            return tipo;
    }
}

export const traduzirMensagemZod = (issue: ZodIssueOptionalMessage): string => {
    switch (issue.code) {
        case ZodIssueCode.invalid_type:
            if(issue.received === "undefined" || issue.received === "null") {
                return "Este campo é obrigatório";
            }
            return `Tipo inválido, deveria ser ${traduzirNomeTipo(issue.expected)} mas recebeu ${traduzirNomeTipo(issue.received)}`;
        case ZodIssueCode.invalid_literal:
            return `Literal inválido, deveria ser ${issue.expected}`;
        case ZodIssueCode.custom:
            return `Erro Desconhecido: ${JSON.stringify(issue.params)}`;
        case ZodIssueCode.invalid_union:
            return `Não é nenhuma das ${issue.unionErrors.length} possibilidades de forma válida`;
        case ZodIssueCode.invalid_union_discriminator:
            return `Deveria ser um desses: ${issue.options?.join(", ")}`;
        case ZodIssueCode.invalid_enum_value:
            return `Deve ser um desses: ${issue.options?.join(", ")}`;
        case ZodIssueCode.unrecognized_keys:
            return `Não esperava estas chaves: ${issue.keys?.join(", ")}`;
        case ZodIssueCode.invalid_arguments:
            return `Argumentos inválidos: ${issue.argumentsError.message}`;
        case ZodIssueCode.invalid_return_type:
            return `Tipo de retorno inválido ${issue.returnTypeError.message}`;
        case ZodIssueCode.invalid_date:
            return "Formato de data inválido.";
        case ZodIssueCode.invalid_string:
            return `Formato de string inválido, ${issue.validation}`;
        case ZodIssueCode.too_small:
            if(issue.type === "number" || issue.type === "bigint")
                return `Deve ser no mínimo ${issue.minimum}`;
            else if(issue.type === "string")
                return `Deve ter no mínimo ${issue.minimum} caracteres`;
            else if(issue.type === "date")
                return `Deve ser após ${new Date(Number(issue.minimum)).toLocaleDateString()}`;
            else
                return `Deve ter no mínimo ${issue.minimum} elementos`;
        case ZodIssueCode.too_big:
            if(issue.type === "number" || issue.type === "bigint")
                return `Deve ser no máximo ${issue.maximum}`;
            else if(issue.type === "string")
                return `Deve ter no máximo ${issue.maximum} caracteres`;
            else if(issue.type === "date")
                return `Deve ser antes de ${new Date(Number(issue.maximum)).toLocaleDateString()}`;
            else
                return `Deve ter no máximo ${issue.maximum} elementos`;
        case ZodIssueCode.invalid_intersection_types:
            return "Tipos de interseção inválidos";
        case ZodIssueCode.not_multiple_of:
            return `O valor não é múltiplo de ${issue.multipleOf}`;
        case ZodIssueCode.not_finite:
            return "O valor não é finito";
        default: 
            return (issue as any).message ?? `Erro Desconhecido: ${JSON.stringify(issue)}`;
    }
};

// https://zod.dev/ERROR_HANDLING
export const customErrorMap: z.ZodErrorMap = (issue, ctx) => {
    return {
        message: traduzirMensagemZod(issue)
    };
};

export const applyZodInitialConfig = () => {
    extendZodWithOpenApi(z);
    
    z.setErrorMap(customErrorMap);
};