import {jest,describe,expect,test, beforeAll, afterAll} from "@jest/globals";
import request  from "supertest";

// Aqui chama conectar banco
import app from "../../src/app.js";

import problemaSeed from "../../src/seeds/problemaSeed.js";
import { ProblemaRepository } from "../../src/models/Problema.js";
import { desconetarBanco } from "../../src/config/dbConnect.js";
import { z } from "zod";
import { ProblemaSchemas } from "../../src/schemas/problemaSchemas.js";
import { CodigoSchemas } from "../../src/schemas/codigoSchemas.js";

describe("Rotas Avaliar Problemas",() => {

	const req = request(app);
    const problemaSchemas = ProblemaSchemas;
    const codigoSchemas = CodigoSchemas;

    let exContarSigla: string | undefined;
    let codigoId: string | undefined;

    beforeAll(async () => {
        await problemaSeed(ProblemaRepository.getInstance());
    });

    test("Deve listar os problemas cadastrados", async () => {
        const response = await req
            .get("/problemas")
            .set("Accept", "aplication/json")
            .expect(200);

        const data = z.array(problemaSchemas.problemaGET).parse(response.body?.data);

        expect(data.length).toBeGreaterThan(0);
    });

    test("Deve criar um problema para contar até um número", async () => {
        const response = await req
            .post("/problemas")
            .set("Accept", "aplication/json")
            .send({
                titulo: "Contar até o valor esperado",
                descricao: "Faça um programa que conte desde 0 até um antes do valor especificado.",
                testes: [
                  {
                    argumentos: ["5"],
                    saida: "0\n1\n2\n3\n4\n",
                    regex: false
                  },
                  {
                    argumentos: ["10"],
                    saida: "0\n1\n2\n3\n4\n5\n6\n7\n8\n9\n",
                    regex: false
                  },
                  {
                    argumentos: ["1"],
                    saida: "0\n",
                    regex: false
                  }
                ],
                testes_privados: [
                  {
                    argumentos: ["20"],
                    saida: "0\n1\n2\n3\n4\n5\n6\n7\n8\n9\n10\n11\n12\n13\n14\n15\n16\n17\n18\n19\n",
                    regex: false
                  }
                ]
              })
            .expect(201);

        const data = problemaSchemas.problemaGET.parse(response.body?.data);
        
        exContarSigla = ""+data.sigla;
    });

    test("Deve obter um problema específico", async () => {
        const response = await req
        .get("/problemas/"+exContarSigla!)
        .set("Accept", "aplication/json")
        .expect(200);

        const data = problemaSchemas.problemaGET.parse(response.body?.data);

        expect(data.sigla).toBe(exContarSigla);

        // Deve ter só os testes públicos
        expect(data.testes.length).toBe(3);

        // Mas a quantidade de testes deve ser a soma dos públicos e privados
        expect(data.testes_quantidade).toBe(4);
    });

    test("Deve enviar um código em JS para ser avaliado", async () => {
        const response = await req
        .post("/problemas/"+exContarSigla!+"/codigo")
        .set("Accept", "aplication/json")
        .send({
            codigo: `
            let limite = parseInt(process.argv[2]);
            for(let i = 0; i < limite; i++) {
                console.log(i);
            }
            `,
            linguagem: "javascript"
        })
        .expect(201);

        const data = codigoSchemas.codigoGET.parse(response.body?.data);

        codigoId = ""+data._id;
    });

    test("Deve verificar que o código foi avaliado", async () => {
        let status: string | undefined;
        let repetiu = 0;
        do {
            const response = await req
            .get("/codigos/"+codigoId!)
            .set("Accept", "aplication/json")
            .expect(200);

            const data = codigoSchemas.codigoGET.parse(response.body?.data);
            status = data.status;
            repetiu++;

            // dorme 1 seg https://stackoverflow.com/questions/45478730/jest-react-testing-check-state-after-delay
            await new Promise((r) => setTimeout(r, 1000));
        } while((status === "Enviado" || status === "Avaliando") && repetiu < 30);

        expect(status).toBe("Sucesso");
    }, 120000);

    afterAll(async () => {
        await desconetarBanco();
    });
});