import { jest, describe, expect, test, afterAll, beforeAll } from "@jest/globals";
import mongoose from "mongoose";

import { MongooseRepository, RepoDoc } from "../../src/utils/mongooseRepository.js";
import { conectarBanco, desconetarBanco } from "../../src/config/dbConnect.js";

describe("MongooseRepository", () => {
    class Teste {
        _id?: mongoose.Types.ObjectId;
        nome: string;
        valores: string[];
        detalhes: { _id?: mongoose.Types.ObjectId, numero: number, data: Date }[];
        extra: { criado?: Date, modificado?: Date };

        constructor(nome: string, valores: string[], detalhes: { numero: number, data: Date }[], extra: { criado?: Date, modificado?: Date }) {
            this.nome = nome;
            this.valores = valores;
            this.detalhes = detalhes;
            this.extra = extra;
        }
    }
    const TesteSchema = new mongoose.Schema<Teste>({
        nome: {
            type: String,
            required: true
        },
        valores: [{
            type: String,
            required: true
        }],
        detalhes: [{
            numero: {
                type: Number,
                required: true
            },
            data: {
                type: Date,
                required: true
            }
        }],
        extra: {
            criado: {
                type: Date,
                required: false
            },
            modificado: {
                type: Date,
                required: false
            }
        }
    });

    let getModel = () => mongoose.model("_testeschemas", TesteSchema);

    let TesteModel: ReturnType<typeof getModel>;
    let repo: MongooseRepository<Teste, typeof TesteModel>;
    let doc: RepoDoc<Teste>;
    let detalhesId: mongoose.Types.ObjectId | undefined;

    beforeAll(async () => {
        await conectarBanco();

        TesteModel = getModel();
        repo = new MongooseRepository<Teste, typeof TesteModel>(TesteModel);
    });

    test("Teste deletar tudo", async () => {
        await repo.deleteEverything();
        
        let quantosDocumentos = await TesteModel.countDocuments({});
        expect(quantosDocumentos).toBe(0);
    });

    test("Teste Criar entrada", async () => {
        let data = {
            nome: "Teste 1",
            valores: ["a", "b", "c"],
            detalhes: [
                { numero: 1, data: new Date() },
                { numero: 2, data: new Date() }
            ],
            extra: {
                criado: new Date()
            }
        };

        let result = await repo.create(data);
        expect(result._id).toBeDefined();
        expect(result).toMatchObject(data);

        doc = result;
    });

    test("Teste Buscar entrada", async () => {
        let result = await repo.findById(doc._id!);
        expect(result === doc).toBe(false);
        expect(result).toBeDefined();
        expect(result!._id!.equals(doc._id)).toBe(true);
    });

    test("Teste patch", async () => {
        let data = {
            valores: ["d", "e", "f"],
            detalhes: [
                { numero: 3, data: new Date() },
                { numero: 4, data: new Date() }
            ],
            extra: {
                modificado: new Date()
            }
        };

        let result = await repo.patch(doc, data);
        expect(result === doc).toBe(true);

        expect(doc.valores).toEqual(["d", "e", "f"]);
        expect(doc.detalhes).toHaveLength(2);
        expect(doc.extra.criado).toBeUndefined();
        expect(doc.extra.modificado).toBeDefined();

        detalhesId = doc.detalhes[0]._id;
        expect(detalhesId).toBeDefined();

        let findAfter = await repo.findById(doc._id!);
        expect(findAfter).toBeTruthy();
        expect(doc.toObject()).toMatchObject(findAfter!.toObject() as any);
    });

    test("Teste patch atualizando um único item em array de objetos", async () => {
        let data_agora = new Date();

        doc.detalhes[0].numero = 100;

        let result = await repo.patch(doc);
        expect(result === doc).toBe(true);

        expect(doc.detalhes).toHaveLength(2);
        expect(doc.detalhes[0]._id).toBe(detalhesId);
        expect(doc.detalhes[0].data < data_agora).toBe(true);
        expect(doc.detalhes[0].numero).toBe(100);

        let findAfter = await repo.findById(doc._id!);
        expect(findAfter).toBeTruthy();
        expect(doc.toObject()).toMatchObject(findAfter!.toObject() as any);
    });

    test("Teste delete", async () => {
        let result = await repo.delete(doc);
        expect(result).toBe(true);

        let otherResult = await repo.findById(doc._id!);
        expect(otherResult).toBeNull();
    });

    test("Remove a collection testes", async () => {
        await mongoose.connection.db!.dropCollection("_testeschemas");
    });

    afterAll(async () => {
        await desconetarBanco();
    });
});