"use server";

import { APIError, IAPIError } from "@/utils/fetchApi";
import { revalidatePath } from "next/cache";
import { cookies } from "next/headers";
import { RedirectType, redirect } from "next/navigation";
import { z } from "zod";

export async function verificaLogin() {
    return (await cookies()).get("token") !== undefined;
}

export async function obterEstadoGlobalInicial() {
    return {
        API_URL: process.env.API_URL_CLIENT,
        ESTA_LOGADO: await verificaLogin(),
        DATA_SERVER: new Date()
    };
}

export async function submitLogin(prevState: unknown, formData: unknown) : Promise<IAPIError | undefined> {
    if(!formData || !(formData instanceof FormData)) {
        return new APIError("FormData não foi passado ou inválido", 400).toJSON();
    }

    const schemaLogin = z.object({
        email: z.string().email(),
        senha: z.string().min(8),
        ref: z.string().default("/"),
    });

    // VALIDAÇÃO DE DADOS
    const result = schemaLogin.safeParse({
        email: formData.get("email"), 
        senha: formData.get("senha"), 
        ref: formData.get("ref")
    });
    
    if(!result.success) {
        let errors = [];
        for (const issue of result.error.issues) {
            errors.push({
                path: issue.path.join("."),
                message: issue.message
            });
        }
        return new APIError(errors, 400).toJSON();
    }

    // Salvar cookie
    (await cookies()).set("token", "ESTÁ LOGADO", {
      httpOnly: true,
      secure: true,
      sameSite: "none"
    });

    // INVALIDAÇÃO DE CACHE
    revalidatePath("/");

    // Redireciona de volta
    redirect(result.data.ref, RedirectType.replace);
}

export async function submitLogout(ref?: unknown) {
    let urlRedirect: string;
    if(typeof ref === "string" && ref) {
        urlRedirect = ref;
    } else {
        urlRedirect = "/";
    }

    // Limpa cookie
    (await cookies()).delete("token");

    // INVALIDAÇÃO DE CACHE
    revalidatePath("/");

    // Redireciona de volta
    redirect(urlRedirect, RedirectType.replace);
}