import { URLSearchParams } from "url";
import { z } from "zod";

export type IAPIError = {
    name: string;
    message: string;
    code: number;
    errors: { message: string, path: string | undefined }[];
};

export class APIError extends Error {

    public code: number;
    public errors: { message: string, path: string | undefined }[];

    constructor(_errors: string|{ message: string, path: string | undefined }|{ message: string, path: string | undefined }[],
        code = 400) {
        super("Erro com código " + code + ": "+_errors);
        this.code = code;

        // unknown aqui para garantir que o tipo de _errors seja discriminado corretamente
        let errors: unknown = _errors;

        if(!errors) {
            this.errors = [{message: "Erro desconhecido", path: undefined}];
        } else if(Array.isArray(errors)) {
            if(errors.length === 0) {
                this.errors = [{message: "Erro desconhecido: []", path: undefined}];
            } else {
                this.errors = errors;
            }
        } else if(typeof errors === "object") {
            this.errors = [{
                message: "message" in errors ? ""+errors.message : "Erro desconhecido",
                path: "path" in errors ? ""+errors.path : undefined
            }];
        } else {
            this.errors = [{message: ""+errors, path: undefined}];
        }
    }

    public toJSON(): IAPIError {
        return {
            name: this.name,
            message: this.message,
            code: this.code,
            errors: this.errors
        };
    }
}

export class APIResponse<T> {
    public code: number;
    public data: T;

    constructor(data: T, code: number = 200) {
        this.code = code;
        this.data = data;
    }

    static build<K extends z.ZodTypeAny>(zodSchema: K, json: any, code: number) {
        const parsed = zodSchema.parse(json) as z.infer<K>;
        return new APIResponse(parsed, code);
    }
}

type HTTPMethod = "GET" | "POST" | "PATCH" |"PUT" | "DELETE" | "HEAD" | "CONNECT" | "OPTIONS";
type HTTPMethodNeedsBody = "POST" | "PATCH" | "PUT";

/**
 * Faz requisições para uma API seguindo o padrão de respostas
 * 
 * @example
 * const resultado = await authFetch("POST", "/auth/local", {
 *    credencial: "usuario",
 *    senha: "senha"
 * });
 */
export const apiFetch = async <T extends z.ZodTypeAny>(zodSchema: T, method: HTTPMethod, route: string, 
    body: Record<string,any> | string | undefined = undefined,
    contentType: string = "application/json") => {

    if(typeof window === "undefined") {
        "use server";
        // https://nextjs.org/docs/app/building-your-application/rendering/server-components#dynamic-functions
        // Isso faz com que qualquer página que chamar fetchServer seja marcada como "force-dynamic"
        // let header = headers();

        if(!process.env.API_URL_SERVER) {
            throw new Error("Tentou fazer fetchServer sem definir env API_URL_SERVER");
        }

        // Se é lado servidor a URL da API tem que ser a do servidor
        return await _apiFetch(zodSchema, process.env.API_URL_SERVER, method, route, body, contentType);
    } else {
        const apiUrl = localStorage.getItem("API_URL");
        if (!apiUrl) {
            throw new Error("Tentou fazer fetchClient sem definir API_URL no localStorage");
        }

        return await _apiFetch(zodSchema, apiUrl, method, route, body, contentType);
    }
};

const _apiFetch = async <T extends z.ZodTypeAny>(zodSchema: T, urlApi: string, method: HTTPMethod, route: string, 
    body: Record<string,any> | string | undefined = undefined,
    contentType: string = "application/json") => {
    let token: string = "?";
    let url: string = urlApi + route;
    let bodyToSend: URLSearchParams | string | undefined = undefined;

    if (!body) {
        // é isso, não precisa fazer nada quando não tem body
    } else if (method === "GET") {
        url += "?" + new URLSearchParams(body);
    } else {
        if (contentType === "application/x-www-form-urlencoded") {
           bodyToSend = new URLSearchParams(body);
        } else if (contentType === "application/json") {
           bodyToSend = JSON.stringify(body);
        } else if (contentType === "text/plain") {
           bodyToSend = body.toString();
        }
    }
   
    console.log("apiFetch", method, url);
    let fetchRes = await fetch(url, {
        method: method,
        //next: {
        //    revalidate: 0
        //},
        headers: {
            "Content-Type": contentType,
            "Authorization": `Bearer ${token}`
        },
        body: bodyToSend,
        cache: "no-store" // next 15
    });

    const code = fetchRes.status;
    // The ok read-only property of the Response interface contains a Boolean stating whether the response was successful (status in the range 200-299) or not.
    const sucess = fetchRes.ok;
    const responseContentType = fetchRes.headers.get("content-type");
    if (responseContentType && responseContentType.includes("application/json")) {
        let json: unknown = await fetchRes.json();

        if (!json || !(typeof json === "object") || Array.isArray(json) || !("data" in json && "errors" in json)) {
            throw new APIError([
                { message: "Resposta em formato inválido", path: undefined },
                { message: ""+json, path: "json" }
            ], code);
        } else if(!sucess || json.errors !== null || json.data === null || !(typeof json.data === "object")) {
            if(json.data === null) {
                throw new APIError(json.errors as any, code);
            } else {
                throw new APIError([
                    { message: "Resposta em formato inválido: foi informado data e errors ao mesmo tempo", path: undefined },
                    { message: ""+json, path: "json" }
                ], code);
            }
        } else {
            return APIResponse.build(zodSchema, json.data, code);
        }
    } else {
        let text: string = await fetchRes.text();

        if(!sucess) {
            throw new APIError(text, code);
        } else {
            return APIResponse.build(zodSchema, text, code);
        }
    }
};


